@@ -18,16 +18,21 @@
 //! [VhostUserBackend]: trait.VhostUserBackend.html
 //! [VhostUserBackendMut]: trait.VhostUserBackendMut.html
 
+use std::fs::File;
 use std::io::Result;
 use std::ops::Deref;
 use std::sync::{Arc, Mutex, RwLock};
 
-use vhost::vhost_user::message::VhostUserProtocolFeatures;
+use vhost::vhost_user::message::{
+    VhostTransferStateDirection, VhostTransferStatePhase, VhostUserProtocolFeatures,
+};
 use vhost::vhost_user::Backend;
 use vm_memory::bitmap::Bitmap;
 use vmm_sys_util::epoll::EventSet;
 use vmm_sys_util::eventfd::EventFd;
 
+use vhost::vhost_user::GpuBackend;
+
 use super::vring::VringT;
 use super::GM;
 
@@ -53,6 +58,12 @@ pub trait VhostUserBackend: Send + Sync {
     /// Get available vhost protocol features.
     fn protocol_features(&self) -> VhostUserProtocolFeatures;
 
+    /// Reset the emulated device state.
+    ///
+    /// A default implementation is provided as we cannot expect all backends to implement this
+    /// function.
+    fn reset_device(&self) {}
+
     /// Enable or disable the virtio EVENT_IDX feature
     fn set_event_idx(&self, enabled: bool);
 
@@ -81,6 +92,18 @@ pub trait VhostUserBackend: Send + Sync {
     /// function.
     fn set_backend_req_fd(&self, _backend: Backend) {}
 
+    /// Set handler for communicating with the frontend by the gpu specific backend communication
+    /// channel.
+    ///
+    /// This function returns a `Result`, returning an error if the backend does not implement this
+    /// function.
+    fn set_gpu_socket(&self, _gpu_backend: GpuBackend) -> Result<()> {
+        Err(std::io::Error::new(
+            std::io::ErrorKind::Unsupported,
+            "backend does not support set_gpu_socket() / VHOST_USER_GPU_SET_SOCKET",
+        ))
+    }
+
     /// Get the map to map queue index to worker thread index.
     ///
     /// A return value of [2, 2, 4] means: the first two queues will be handled by worker thread 0,
@@ -110,6 +133,37 @@ pub trait VhostUserBackend: Send + Sync {
         vrings: &[Self::Vring],
         thread_id: usize,
     ) -> Result<()>;
+
+    /// Initiate transfer of internal state for the purpose of migration to/from the back-end.
+    ///
+    /// Depending on `direction`, the state should either be saved (i.e. serialized and written to
+    /// `file`) or loaded (i.e. read from `file` and deserialized). The back-end can choose to use
+    /// a different channel than file. If so, it must return a File that the front-end can use.
+    /// Note that this function must not block during transfer, i.e. I/O to/from `file` must be
+    /// done outside of this function.
+    fn set_device_state_fd(
+        &self,
+        _direction: VhostTransferStateDirection,
+        _phase: VhostTransferStatePhase,
+        _file: File,
+    ) -> Result<Option<File>> {
+        Err(std::io::Error::new(
+            std::io::ErrorKind::Unsupported,
+            "back end does not support state transfer",
+        ))
+    }
+
+    /// After transferring internal state, check for any resulting errors, including potential
+    /// deserialization errors when loading state.
+    ///
+    /// Although this function return a `Result`, the front-end will not receive any details about
+    /// this error.
+    fn check_device_state(&self) -> Result<()> {
+        Err(std::io::Error::new(
+            std::io::ErrorKind::Unsupported,
+            "back end does not support state transfer",
+        ))
+    }
 }
 
 /// Trait without interior mutability for vhost user backend servers to implement concrete services.
@@ -132,6 +186,12 @@ pub trait VhostUserBackendMut: Send + Sync {
     /// Get available vhost protocol features.
     fn protocol_features(&self) -> VhostUserProtocolFeatures;
 
+    /// Reset the emulated device state.
+    ///
+    /// A default implementation is provided as we cannot expect all backends to implement this
+    /// function.
+    fn reset_device(&mut self) {}
+
     /// Enable or disable the virtio EVENT_IDX feature
     fn set_event_idx(&mut self, enabled: bool);
 
@@ -160,6 +220,18 @@ pub trait VhostUserBackendMut: Send + Sync {
     /// function.
     fn set_backend_req_fd(&mut self, _backend: Backend) {}
 
+    /// Set handler for communicating with the frontend by the gpu specific backend communication
+    /// channel.
+    ///
+    /// This function returns a `Result`, returning an error if the backend does not implement this
+    /// function.
+    fn set_gpu_socket(&mut self, _gpu_backend: GpuBackend) -> Result<()> {
+        Err(std::io::Error::new(
+            std::io::ErrorKind::Unsupported,
+            "backend does not support set_gpu_socket() / VHOST_USER_GPU_SET_SOCKET",
+        ))
+    }
+
     /// Get the map to map queue index to worker thread index.
     ///
     /// A return value of [2, 2, 4] means: the first two queues will be handled by worker thread 0,
@@ -190,6 +262,35 @@ pub trait VhostUserBackendMut: Send + Sync {
         vrings: &[Self::Vring],
         thread_id: usize,
     ) -> Result<()>;
+
+    /// Initiate transfer of internal state for the purpose of migration to/from the back-end.
+    ///
+    /// Depending on `direction`, the state should either be saved (i.e. serialized and written to
+    /// `file`) or loaded (i.e. read from `file` and deserialized).  Note that this function must
+    /// not block during transfer, i.e. I/O to/from `file` must be done outside of this function.
+    fn set_device_state_fd(
+        &mut self,
+        _direction: VhostTransferStateDirection,
+        _phase: VhostTransferStatePhase,
+        _file: File,
+    ) -> Result<Option<File>> {
+        Err(std::io::Error::new(
+            std::io::ErrorKind::Unsupported,
+            "back end does not support state transfer",
+        ))
+    }
+
+    /// After transferring internal state, check for any resulting errors, including potential
+    /// deserialization errors when loading state.
+    ///
+    /// Although this function return a `Result`, the front-end will not receive any details about
+    /// this error.
+    fn check_device_state(&self) -> Result<()> {
+        Err(std::io::Error::new(
+            std::io::ErrorKind::Unsupported,
+            "back end does not support state transfer",
+        ))
+    }
 }
 
 impl<T: VhostUserBackend> VhostUserBackend for Arc<T> {
@@ -216,6 +317,10 @@ impl<T: VhostUserBackend> VhostUserBackend for Arc<T> {
         self.deref().protocol_features()
     }
 
+    fn reset_device(&self) {
+        self.deref().reset_device()
+    }
+
     fn set_event_idx(&self, enabled: bool) {
         self.deref().set_event_idx(enabled)
     }
@@ -236,6 +341,10 @@ impl<T: VhostUserBackend> VhostUserBackend for Arc<T> {
         self.deref().set_backend_req_fd(backend)
     }
 
+    fn set_gpu_socket(&self, gpu_backend: GpuBackend) -> Result<()> {
+        self.deref().set_gpu_socket(gpu_backend)
+    }
+
     fn queues_per_thread(&self) -> Vec<u64> {
         self.deref().queues_per_thread()
     }
@@ -254,6 +363,19 @@ impl<T: VhostUserBackend> VhostUserBackend for Arc<T> {
         self.deref()
             .handle_event(device_event, evset, vrings, thread_id)
     }
+
+    fn set_device_state_fd(
+        &self,
+        direction: VhostTransferStateDirection,
+        phase: VhostTransferStatePhase,
+        file: File,
+    ) -> Result<Option<File>> {
+        self.deref().set_device_state_fd(direction, phase, file)
+    }
+
+    fn check_device_state(&self) -> Result<()> {
+        self.deref().check_device_state()
+    }
 }
 
 impl<T: VhostUserBackendMut> VhostUserBackend for Mutex<T> {
@@ -280,6 +402,10 @@ impl<T: VhostUserBackendMut> VhostUserBackend for Mutex<T> {
         self.lock().unwrap().protocol_features()
     }
 
+    fn reset_device(&self) {
+        self.lock().unwrap().reset_device()
+    }
+
     fn set_event_idx(&self, enabled: bool) {
         self.lock().unwrap().set_event_idx(enabled)
     }
@@ -300,6 +426,10 @@ impl<T: VhostUserBackendMut> VhostUserBackend for Mutex<T> {
         self.lock().unwrap().set_backend_req_fd(backend)
     }
 
+    fn set_gpu_socket(&self, gpu_backend: GpuBackend) -> Result<()> {
+        self.lock().unwrap().set_gpu_socket(gpu_backend)
+    }
+
     fn queues_per_thread(&self) -> Vec<u64> {
         self.lock().unwrap().queues_per_thread()
     }
@@ -319,6 +449,21 @@ impl<T: VhostUserBackendMut> VhostUserBackend for Mutex<T> {
             .unwrap()
             .handle_event(device_event, evset, vrings, thread_id)
     }
+
+    fn set_device_state_fd(
+        &self,
+        direction: VhostTransferStateDirection,
+        phase: VhostTransferStatePhase,
+        file: File,
+    ) -> Result<Option<File>> {
+        self.lock()
+            .unwrap()
+            .set_device_state_fd(direction, phase, file)
+    }
+
+    fn check_device_state(&self) -> Result<()> {
+        self.lock().unwrap().check_device_state()
+    }
 }
 
 impl<T: VhostUserBackendMut> VhostUserBackend for RwLock<T> {
@@ -345,6 +490,10 @@ impl<T: VhostUserBackendMut> VhostUserBackend for RwLock<T> {
         self.read().unwrap().protocol_features()
     }
 
+    fn reset_device(&self) {
+        self.write().unwrap().reset_device()
+    }
+
     fn set_event_idx(&self, enabled: bool) {
         self.write().unwrap().set_event_idx(enabled)
     }
@@ -365,6 +514,10 @@ impl<T: VhostUserBackendMut> VhostUserBackend for RwLock<T> {
         self.write().unwrap().set_backend_req_fd(backend)
     }
 
+    fn set_gpu_socket(&self, gpu_backend: GpuBackend) -> Result<()> {
+        self.write().unwrap().set_gpu_socket(gpu_backend)
+    }
+
     fn queues_per_thread(&self) -> Vec<u64> {
         self.read().unwrap().queues_per_thread()
     }
@@ -384,6 +537,21 @@ impl<T: VhostUserBackendMut> VhostUserBackend for RwLock<T> {
             .unwrap()
             .handle_event(device_event, evset, vrings, thread_id)
     }
+
+    fn set_device_state_fd(
+        &self,
+        direction: VhostTransferStateDirection,
+        phase: VhostTransferStatePhase,
+        file: File,
+    ) -> Result<Option<File>> {
+        self.write()
+            .unwrap()
+            .set_device_state_fd(direction, phase, file)
+    }
+
+    fn check_device_state(&self) -> Result<()> {
+        self.read().unwrap().check_device_state()
+    }
 }
 
 #[cfg(test)]
@@ -445,6 +613,12 @@ pub mod tests {
             VhostUserProtocolFeatures::all()
         }
 
+        fn reset_device(&mut self) {
+            self.event_idx = false;
+            self.events = 0;
+            self.acked_features = 0;
+        }
+
         fn set_event_idx(&mut self, enabled: bool) {
             self.event_idx = enabled;
         }
@@ -524,6 +698,11 @@ pub mod tests {
             GuestMemoryMmap::<()>::from_ranges(&[(GuestAddress(0x100000), 0x10000)]).unwrap(),
         );
         backend.update_memory(mem).unwrap();
+
+        backend.reset_device();
+        assert!(backend.lock().unwrap().events == 0);
+        assert!(!backend.lock().unwrap().event_idx);
+        assert!(backend.lock().unwrap().acked_features == 0);
     }
 
     #[test]
@@ -559,5 +738,10 @@ pub mod tests {
         backend
             .handle_event(0x1, EventSet::IN, &[vring], 0)
             .unwrap();
+
+        backend.reset_device();
+        assert!(backend.read().unwrap().events == 0);
+        assert!(!backend.read().unwrap().event_idx);
+        assert!(backend.read().unwrap().acked_features == 0);
     }
 }
